unit BCGI;

{$mode objfpc}{$H+}

interface

uses
{$IFDEF CGIDEBUG}
  DbugIntf,
{$ENDIF}
  BrookApplication, BrookException, BrookWebHandler, BrookMessages, BrookConsts,
  BrookHTTPConsts, HTTPDefs, CustWeb, CustCGI, Classes, SysUtils;

type
  TBCGIApplication = class;

  { TBApplication }

  TBApplication = class(TInterfacedObject, IBrookApplication)
  private
    FApp: TBCGIApplication;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function Instance: TCustomWebApplication;
    procedure Run;
  end;

  { TBCGIApplication }

  TBCGIApplication = class(TCustomCGIApplication)
  protected
    function InitializeWebHandler: TWebHandler; override;
  end;

  { TBCGIRequest }

  TBCGIRequest = class(TCGIRequest)
  protected
    procedure InitRequestVars; override;
    procedure InitCustomVars; virtual;
  end;

  { TBCGIResponse }

  TBCGIResponse = class(TCGIResponse)
  protected
    procedure CollectHeaders(AHeaders: TStrings); override;
  end;

  { TBCGIHandler }

  TBCGIHandler = class(TCGIHandler)
  protected
    function CreateRequest: TCGIRequest; override;
    function CreateResponse(AOutput: TStream): TCGIResponse; override;
  public
    function FormatContentType(const AContentType: string): string;
    procedure HandleRequest(ARequest: TRequest; AResponse: TResponse); override;
    procedure ShowRequestException(R: TResponse; E: Exception); override;
  end;

procedure SetDefaultCharset(const ACharset: ShortString);
procedure SetPage404(const APage: string);
procedure SetPage500(const APage: string);

implementation

var
  _DefaultCharset: ShortString = ES;
  _Page404: string = ES;
  _Page500: string = ES;

procedure SetDefaultCharset(const ACharset: ShortString);
begin
  _DefaultCharset := ACharset;
end;

procedure SetPage404(const APage: string);
begin
  _Page404 := APage;
end;

procedure SetPage500(const APage: string);
begin
  _Page500 := APage;
end;

{ TBCGIRequest }

procedure TBCGIRequest.InitRequestVars;
var
  VMethod: ShortString;
begin
{$IFDEF CGIDEBUG}
  SendMethodEnter('TBCGIRequest.InitRequestVars');
{$ENDIF}
  VMethod := Method;
  if VMethod = ES then
    raise Exception.Create(SNoRequestMethodError);
  if (CompareText(VMethod, HTTP_REQUEST_METHOD_POST) = 0) or
    (CompareText(VMethod, HTTP_REQUEST_METHOD_PUT) = 0) then
  begin
    InitPostVars;
    if HandleGetOnPost then
      InitGetVars;
  end
  else
  if (CompareText(VMethod, HTTP_REQUEST_METHOD_GET) = 0) or
    (CompareText(VMethod, HTTP_REQUEST_METHOD_DELETE) = 0) or
    (CompareText(VMethod, HTTP_REQUEST_METHOD_OPTIONS) = 0) then
    InitCustomVars;
{$IFDEF CGIDEBUG}
  SendMethodExit('TBCGIRequest.InitRequestVars');
{$ENDIF}
end;

procedure TBCGIRequest.InitCustomVars;
begin
end;

{ TBCGIResponse }

procedure TBCGIResponse.CollectHeaders(AHeaders: TStrings);
begin
  AHeaders.Add(HTTP_HEADER_X_POWERED_BY + HS + 'Brook framework and B-CGI plugin.');
  inherited CollectHeaders(AHeaders);
end;

{ TBApplication }

constructor TBApplication.Create;
begin
  FApp := TBCGIApplication.Create(nil);
  FApp.Initialize;
end;

destructor TBApplication.Destroy;
begin
  FApp.Free;
  inherited Destroy;
end;

function TBApplication.Instance: TCustomWebApplication;
begin
  Result := FApp;
end;

procedure TBApplication.Run;
begin
  FApp.Run;
end;

{ TBCGIApplication }

function TBCGIApplication.InitializeWebHandler: TWebHandler;
begin
  Result := TBCGIHandler.Create(Self);
end;

{ TBCGIHandler }

function TBCGIHandler.CreateRequest: TCGIRequest;
begin
  Result := TBCGIRequest.CreateCGI(Self);
  if ApplicationURL = ES then
    ApplicationURL := Result.ScriptName;
end;

function TBCGIHandler.CreateResponse(AOutput: TStream): TCGIResponse;
begin
  Result := TBCGIResponse.CreateCGI(Self, AOutput);
end;

function TBCGIHandler.FormatContentType(const AContentType: string): string;
begin
  if _DefaultCharset <> ES then
    Result := AContentType + HTTP_HEADER_CHARSET + _DefaultCharset
  else
    Result := AContentType;
end;

procedure TBCGIHandler.HandleRequest(ARequest: TRequest; AResponse: TResponse);
begin
  AResponse.ContentType := FormatContentType(AResponse.ContentType);
  TBrookWebHandler.DefaultService.HandleRequest(ARequest, AResponse);
end;

procedure TBCGIHandler.ShowRequestException(R: TResponse; E: Exception);
var
  VStr: TStrings;
  VHandled: Boolean = False;
begin
  if R.ContentSent then
    Exit;
  if Assigned(OnShowRequestException) then
  begin
    OnShowRequestException(R, E, VHandled);
    if VHandled then
      Exit;
  end;
  if RedirectOnError and not R.HeadersSent then
  begin
    R.SendRedirect(Format(RedirectOnErrorURL, [HTTPEncode(E.Message)]));
    R.SendContent;
    Exit;
  end;
  if (_Page404 <> ES) and (E is EBrookHTTP404) and (not R.HeadersSent) then
  begin
    R.Code := HTTP_STATUS_CODE_NOT_FOUND;
    R.CodeText := HTTP_REASON_PHRASE_NOT_FOUND;
    R.ContentType := FormatContentType(HTTP_CONTENT_TYPE_TEXT_HTML);
    if FileExists(_Page404) then
      R.Contents.LoadFromFile(_Page404)
    else
      R.Content := _Page404;
    R.Content := Format(R.Content, [ApplicationURL]);
    R.SendContent;
    Exit;
  end;
  if (_Page500 <> ES) and (not R.HeadersSent) then
  begin
    R.Code := HTTP_STATUS_CODE_INTERNAL_SERVER_ERROR;
    R.CodeText := 'Application error ' + E.ClassName;
    R.ContentType := FormatContentType(HTTP_CONTENT_TYPE_TEXT_HTML);
    if FileExists(_Page500) then
      R.Contents.LoadFromFile(_Page500)
    else
      R.Content := _Page500;
    R.Content := Format(R.Content, [E.Message]);
    R.SendContent;
    Exit;
  end;
  if R.ContentType = HTTP_CONTENT_TYPE_TEXT_HTML then
  begin
    VStr := TStringList.Create;
    try
      R.ContentType := FormatContentType(HTTP_CONTENT_TYPE_TEXT_HTML);
      ExceptionToHTML(VStr, E, Title, Email, Administrator);
      R.Content := VStr.Text;
      R.SendContent;
    finally
      VStr.Free;
    end;
  end;
end;

initialization
  BrookRegisterApp(TBApplication.Create);

end.
